/*
 * Maix Bit I2S player
 *   SD card : "*.wav"
 *       Linear PCM, 8-22.05kHz, Signed 16bit, Stereo
 *   I2S : BCLK 35
 *         LRC  33
 *         DOUT 34
 */
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "i2s.h"
#include "sysctl.h"
#include "fpioa.h"
#include "uarths.h"

#define RX_BUF_THRESHOLD     (1024*8)
#define MAX_BUF_LEN          (1024*16)

#include <SPI.h>
#include <SD.h>
#include <Sipeed_ST7789_game.h>

#define SD_CS 29
#define LED_PIN 12 // LED G

SPIClass spi_(SPI0); // MUST be SPI0 for Maix series on board LCD
Sipeed_ST7789 tft(320, 240, spi_);
#define MAX_LINE 26
int _width, _height;

#define MAX_FILE 128
File root;
char fname[MAX_FILE][32];
int fnum;

#define VOL_MAX 16
int  num, old_num;
int  disp_num;
int  vol, old_vol;

int  vol256[VOL_MAX+1]={
    0,  1,  2,  3,  4,  6,   8,  11,
   16, 23, 32, 45, 64, 90, 128, 181,
  256
};

int StickX0Pin = 24;
int StickX1Pin = 25;
int StickY0Pin = 22;
int StickY1Pin = 23;
int StickX0;
int StickX1;
int StickY0;
int StickY1;
int OldX0;
int OldX1;
int OldY0;
int OldY1;

int ButtonAPin = 30;
int ButtonBPin = 31;
int ButtonA;
int ButtonB;
int OldA;
int OldB;

int ftype;
int I2S_stop;

#define BUF_SIZE 4096
const int offset = 0x2C;
const uint8_t *data0;
const uint8_t *data1;

void setup() {
  Serial.begin(115200);
  
  tft.begin(15000000);
  tft.setRotation(0);
//  tft.invertDisplay(true);
  _width=tft.width();
  _height=tft.height();

  Serial.print("Initializing SD card...");
//  pinMode(SD_CS, PULLUP);
//  pinMode(MISO, PULLUP);
  if (!SD.begin(40000000, SD_CS)) {
    Serial.println("failed!");
    show_no_sd();
    while(1);
  }
  root = SD.open("/");
  pinMode(LED_PIN, OUTPUT);

  tft.fillRect(0, 0, _width, _height, COLOR_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(COLOR_WHITE, COLOR_BLACK);
  tft.setCursor(_width/2-64, _height/2);
  tft.print("Now read SD card");
  tft.display();

  fnum=0;
  printDirectory(root, 0, ".wav");

  Serial.println("OK!");

  uarths_init();
  fpioa_set_function(34, FUNC_I2S0_OUT_D1);
  fpioa_set_function(35, FUNC_I2S0_SCLK);
  fpioa_set_function(33, FUNC_I2S0_WS);

  pinMode(StickX0Pin, INPUT_PULLUP);
  pinMode(StickX1Pin, INPUT_PULLUP);
  pinMode(StickY0Pin, INPUT_PULLUP);
  pinMode(StickY1Pin, INPUT_PULLUP);
  pinMode(ButtonAPin, INPUT_PULLUP);
  pinMode(ButtonBPin, INPUT_PULLUP);

  ftype=0;
  I2S_stop=0;
  num=0;
  old_num=-1;
  disp_num=0;
  vol=8;
  data0=(const uint8_t *)malloc(BUF_SIZE);
  data1=(const uint8_t *)malloc(BUF_SIZE);
}

void loop() {
  int i;
  int push_btn;
  char str[16];
  
  digitalWrite(LED_PIN, 0);
  OldX0=StickX0;
  OldX1=StickX1;
  OldY0=StickY0;
  OldY1=StickY1;
  OldA=ButtonA;
  OldB=ButtonB;
  StickX0 = digitalRead(StickX0Pin);
  StickX1 = digitalRead(StickX1Pin);
  StickY0 = digitalRead(StickY0Pin);
  StickY1 = digitalRead(StickY1Pin);
  ButtonA = digitalRead(ButtonAPin);
  ButtonB = digitalRead(ButtonBPin);

  if (ftype==0){
    I2S_stop=0;
    if (StickX0==0){
      if (vol>0) vol--;
      delay(200);
    }
    if (StickX1==0){
      if (vol<VOL_MAX) vol++;
      delay(200);
    }
    
    if (StickY0==0){
      num++;
      delay(200);
    }
    if (StickY1==0){
      num--;
      delay(200);
    }
    if (num<0)
      num=fnum-1;
    if (num>=fnum)
      num=0;

    if (ButtonA==0 && OldA==1){
      if (num==fnum)
        ftype=2;
      else
        ftype=1;
    }

    if (ButtonB==0 && OldB==1){
      ftype=0;
      num=0;
      old_num=-1;
    }

    if (num==0)
      disp_num=0;
    else if (num==fnum && fnum>=MAX_LINE)
      disp_num=num-MAX_LINE+1;
    else if (num-disp_num>=(MAX_LINE-1))
      disp_num=num-MAX_LINE+2;
    else if (num-disp_num<=0)
      disp_num=num-1;

    if (num!=old_num)
      show_wav_sel();

    if (vol!=old_vol){
      tft.setTextSize(1);
      tft.setTextColor(COLOR_WHITE, COLOR_BLACK);
      tft.setCursor(0, _height-9);
      sprintf(str, "Volume : %d  ", vol);
      tft.print(str);
      tft.display();
    }
    old_num=num;
    old_vol=vol;

  }else if (ftype==1 || ftype==2){
    I2Splay(fname[num]);
    if (I2S_stop>0){
      ftype=0;
      old_num=-1;
    }else{
      num++;
      if (num>=fnum)
        num=0;
    }
  }else if (ftype==2){
    for(i=0;i<fnum;i++){
      I2Splay(fname[i]);
      if (I2S_stop>0)
        break;
    }
    ftype=0;
    old_num=-1;
  }
  delay(50);
}

void printDirectory(File dir, int numTabs, char *typestr) {
  while (true) {

    File entry =  dir.openNextFile();
    if (! entry) {
      // no more files
      break;
    }
    for (uint8_t i = 0; i < numTabs; i++) {
      Serial.print('\t');
    }
    Serial.print(entry.name());
    if (fnum<MAX_FILE && stristr(entry.name(), typestr)!=NULL){
      if (numTabs>0){
        strcpy((char *)fname[fnum], (char *)dir.name());
        strcat((char *)fname[fnum], "/");
      }else
        fname[fnum][0]='\0';
      strcat((char *)fname[fnum], (char *)entry.name());
      fnum++;
    }else if(fnum>=MAX_FILE){
      entry.close();
      return;
    }
    
    if (entry.isDirectory()) {
      Serial.println("/");
      printDirectory(entry, numTabs + 1, typestr);
    } else {
      // files have sizes, directories do not
      Serial.print("\t\t");
      Serial.println(entry.size(), DEC);
    }
    entry.close();
  }
}

uint16_t read16(File &f) {
  uint16_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read(); // MSB
  return result;
}

uint32_t read32(File &f) {
  uint32_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read();
  ((uint8_t *)&result)[2] = f.read();
  ((uint8_t *)&result)[3] = f.read(); // MSB
  return result;
}

char *stristr(const char *org, char *str)
{
  char *tmp, *tmp0, *tmp1;
  tmp=(char *)org;
  for(;*tmp!='\0';tmp++){
    tmp0=tmp;
    tmp1=str;
    for(;*tmp1!='\0';tmp1++){
      if (*tmp0==*tmp1){
        tmp0++;
        continue;
      }else if ('A'<=*tmp1 && *tmp1<='Z' && *tmp0==(*tmp1-'A'+'a')){
        tmp0++;
        continue;
      }else if ('a'<=*tmp1 && *tmp1<='z' && *tmp0==(*tmp1-'a'+'A')){
        tmp0++;
        continue;
      }else
        break;
    }
    if (*tmp1=='\0')
      return(tmp);
  }
  return(NULL);
}

void show_no_sd()
{
  tft.fillRect(0, 0, _width, _height, COLOR_BLACK);
  tft.setTextSize(2);
  tft.setTextColor(COLOR_RED, COLOR_BLACK);
  tft.setCursor(16, 0);
  tft.print("Error");

  tft.setTextSize(2);
  tft.setTextColor(COLOR_WHITE, COLOR_BLACK);
  tft.setCursor(0, 32);
  tft.print("No SD card");
  tft.display();
}

void show_wav_sel()
{
  int i;
  int j;
  
  tft.fillRect(0, 0, _width, _height-9, COLOR_BLACK);
  tft.setTextSize(2);
  tft.setTextColor(COLOR_GREEN);
  tft.setCursor(0, 0);
  tft.print("WAV(PCM 16bit) file");

  tft.setTextSize(1);
  tft.setTextColor(COLOR_WHITE);
  if (fnum<=0){
    tft.setCursor(0, 16);
    tft.print("No data");
    tft.display();
    return;
  }

  tft.fillRect(0, 16+(num-disp_num)*8, _width, 8, COLOR_RED);
  for(i=0;i+disp_num<fnum && i<MAX_LINE;i++){
    j=i+disp_num;
    tft.setCursor(0, 16+i*8);
    tft.print(fname[j]);
  }
  tft.display();
}

void I2Splay(char *fname)
{
  int i;
  int sz;
  int val;
  int wav_rate;
  int16_t *tmp;
  uint8_t wav_header[offset];
  i2s_data_t i2s_data;
  char str[16];
  
  tft.fillRect(0, 0, _width, _height-9, COLOR_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(COLOR_WHITE, COLOR_BLACK);
  tft.setCursor(0, _height/2-16);
  tft.print(fname);
  tft.print("                                ");
  tft.display();

  File file = SD.open(fname);  // 8-44.1KHz, 16bit, stereo, linear PCM
  file.readBytes(wav_header, offset);
  wav_rate=(int)wav_header[24]+(int)wav_header[25]*256;
  i2s_init(I2S_DEVICE_0, I2S_TRANSMITTER, 0x0C);
  i2s_tx_channel_config(I2S_DEVICE_0, I2S_CHANNEL_1,
      RESOLUTION_16_BIT, SCLK_CYCLES_32,
      TRIGGER_LEVEL_4,
      RIGHT_JUSTIFYING_MODE
  );
  Serial.println(fname);
  Serial.print("Rate : ");
  Serial.println(wav_rate);
  i2s_set_sample_rate(I2S_DEVICE_0, wav_rate);

  while (1){
    OldX0=StickX0;
    OldX1=StickX1;
    OldB=ButtonB;
    StickX0 = digitalRead(StickX0Pin);
    StickX1 = digitalRead(StickX1Pin);
    ButtonB = digitalRead(ButtonBPin);
//    if (StickX0==0 && OldX0==1) {
    if (StickX0==0) {
      if (vol>0) vol--;
      delay(200);
    }
//    if (StickX1==0 && OldX1==1) {
    if (StickX1==0) {
      if (vol<VOL_MAX) vol++;
      delay(200);
    }
    if (ButtonB==0 && OldB==1){
      I2S_stop=1;
    }

    digitalWrite(LED_PIN, 1);
    sz=file.readBytes((char *)data0, BUF_SIZE);
    digitalWrite(LED_PIN, 0);
    if (sz==0 || I2S_stop==1)
      break;
    tmp=(int16_t *)data0;
    for(i=0; i<sz; i+=2){
      val=(int)*tmp;
      val=val*vol256[vol]/256;
      *(tmp++)=(int16_t)val;
    }
    i2s_play(I2S_DEVICE_0, DMAC_CHANNEL0,
      data0, sz, 1024, 16, 2);

    if (vol!=old_vol){
      tft.setTextSize(1);
      tft.setTextColor(COLOR_WHITE, COLOR_BLACK);
      tft.setCursor(0, _height-9);
      sprintf(str, "Volume : %d  ", vol);
      tft.print(str);
      tft.display();
    }
    old_vol=vol;

    digitalWrite(LED_PIN, 1);
    sz=file.readBytes((char *)data1, BUF_SIZE);
    digitalWrite(LED_PIN, 0);
    if (sz==0 || I2S_stop==1)
      break;
    tmp=(int16_t *)data1;
    for(i=0; i<sz; i+=2){
      val=(int)*tmp;
      val=val*vol256[vol]/256;
      *(tmp++)=(int16_t)val;
    }
    i2s_play(I2S_DEVICE_0, DMAC_CHANNEL0,
      data1, sz, 1024, 16, 2);
  }
  file.close();
  digitalWrite(LED_PIN, 0);

  // Sound stop
  tmp=(int16_t *)data0;
  for(i=0; i<BUF_SIZE; i+=2)
    *(tmp++)=0;
  tmp=(int16_t *)data1;
  for(i=0; i<BUF_SIZE; i+=2)
    *(tmp++)=0;
  i2s_play(I2S_DEVICE_0, DMAC_CHANNEL0,
    data0, BUF_SIZE, 1024, 16, 2);
}
