
#include "Sipeed_ST7789_game.h"
#include "lcd.h"
#include "sysctl.h"
#include "utils.h"

Sipeed_ST7789::Sipeed_ST7789(uint16_t w, uint16_t h, SPIClass& spi, int8_t dc_pin, int8_t rst_pin, uint8_t dma_ch )
:Adafruit_GFX(w,h),
 _spi(spi), _dcxPin(dc_pin), _rstPin(rst_pin), 
 _dmaCh(dma_ch),
 _screenDir(DIR_YX_RLDU)
{
    configASSERT(_spi.busId()==SPI0);
}

Sipeed_ST7789::~Sipeed_ST7789(void)
{

}

boolean Sipeed_ST7789::begin( uint32_t freq, uint16_t color )
{
    uint8_t spiNum = _spi.busId();
    if( (spi_id_t)spiNum == SPI0)
    {
        fpioa_set_function(SIPEED_ST7789_SS_PIN  , (fpioa_function_t)(FUNC_SPI0_SS0 + SIPEED_ST7789_SS));
        fpioa_set_function(SIPEED_ST7789_SCLK_PIN, (fpioa_function_t)FUNC_SPI0_SCLK);
    }
    else if((spi_id_t)spiNum == SPI1)
    {
        fpioa_set_function(SIPEED_ST7789_SS_PIN  , (fpioa_function_t)(FUNC_SPI1_SS0 + SIPEED_ST7789_SS));
        fpioa_set_function(SIPEED_ST7789_SCLK_PIN, (fpioa_function_t)FUNC_SPI1_SCLK);
    }
    sysctl_set_spi0_dvp_data(1);
    _freq = freq;

    buffer[0] = (uint8_t *)malloc(LCD_X_MAX * LCD_Y_MAX * 2 / 4);
    buffer[1] = (uint8_t *)malloc(LCD_X_MAX * LCD_Y_MAX * 2 / 4);
    buffer[2] = (uint8_t *)malloc(LCD_X_MAX * LCD_Y_MAX * 2 / 4);
    buffer[3] = (uint8_t *)malloc(LCD_X_MAX * LCD_Y_MAX * 2 / 4);

    lcd_init(spiNum, SIPEED_ST7789_SS, SIPEED_ST7789_RST_GPIONUM, SIPEED_ST7789_DCX_GPIONUM, _freq, _rstPin,  _dcxPin, _dmaCh);
    lcd_set_direction((lcd_dir_t)_screenDir);
    lcd_clear(color);
    return true;
}

void Sipeed_ST7789::display()
{
    lcd_set_area(0, 0, _width, _height / 4);
    tft_write_half((uint16_t*)buffer[0], LCD_X_MAX * LCD_Y_MAX / 4);
    lcd_set_area(0, _height / 4, _width, _height*2 / 4);
    tft_write_half((uint16_t*)buffer[1], LCD_X_MAX * LCD_Y_MAX / 4);
    lcd_set_area(0, _height*2 / 4, _width, _height*3 / 4);
    tft_write_half((uint16_t*)buffer[2], LCD_X_MAX * LCD_Y_MAX / 4);
    lcd_set_area(0, _height*3 / 4, _width, _height);
    tft_write_half((uint16_t*)buffer[3], LCD_X_MAX * LCD_Y_MAX / 4);
}
void Sipeed_ST7789::drawPixel(int16_t x, int16_t y, uint16_t color)
{
    uint16_t i;
    uint16_t m, n;
    uint8_t *tmppt;

    if (x<0 || x>=_width || y<0 || y>=_height)
        return;

    m=y * 4 / _height;
    n=y - m * _height / 4;

    i=((n * _width) + x) * 2;
    tmppt=&buffer[m][i];
    *(tmppt++)=color & 0xFF;
    *tmppt=color >> 8;
}
void Sipeed_ST7789::writePixel(int16_t x, int16_t y, uint16_t color)
{
    drawPixel(x, y, color);
}

void Sipeed_ST7789::writeFillRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color)
{
    uint16_t i;
    uint16_t m, n;
    int16_t  x0, y0;
    int16_t  x1, y1;
    int16_t  px, py;
    uint8_t *tmppt;

    if ((x+w)<0 || x>=_width || (y+h)<0 || y>=_height)
        return;

    x0=x;
    x1=x+w;
    y0=y;
    y1=y+h;

    if (x<0)
      x0=0;
    if ((x+w)>_width)
      x1=_width;

    if (y<0)
      y0=0;
    if ((y+h)>_height)
      y1=_height;

    for(py=y0; py<y1; py++){
      m=py * 4 / _height;
      n=py - m * _height / 4;
      for(px=x0; px<x1; px++){
        i=((n * _width) + px) * 2;
        tmppt=&buffer[m][i];
        *(tmppt++)=color & 0xFF;
        *tmppt=color >> 8;
      }
    }
}

void Sipeed_ST7789::writeFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color)
{
    uint16_t i;
    uint16_t m, n;
    int16_t  y0, y1;
    int16_t  px, py;
    uint8_t *tmppt;

    if (x<0 || x>=_width || (y+h)<0 || y>=_height)
        return;

    y0=y;
    y1=y+h;

    if (y<0)
      y0=0;
    if ((y+h)>_height)
      y1=_height;

    px=x;
    for(py=y0; py<y1; py++){
        m=py * 4 / _height;
        n=py - m * _height / 4;

        i=((n * _width) + px) * 2;
        tmppt=&buffer[m][i];
        *(tmppt++)=color & 0xFF;
        *tmppt=color >> 8;
    }
}

void Sipeed_ST7789::writeFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color)
{
    uint16_t i;
    uint16_t m, n;
    int16_t  x0, x1;
    int16_t  px, py;
    uint8_t *tmppt;

    if ((x+w)<0 || x>=_width || y<0 || y>=_height)
        return;

    x0=x;
    x1=x+w;

    if (x<0)
      x0=0;
    if ((x+w)>_width)
      x1=_width;

    py=y;
    m=py * 4 / _height;
    n=py - m * _height / 4;
    for(px=x0; px<x1; px++){
        i=((n * _width) + px) * 2;
        tmppt=&buffer[m][i];
        *(tmppt++)=color & 0xFF;
        *tmppt=color >> 8;
    }
}

void Sipeed_ST7789::drawFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color)
{
    writeFastVLine(x, y, h, color);
}

void Sipeed_ST7789::drawFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color)
{
    writeFastHLine(x, y, w, color);
}

void Sipeed_ST7789::fillRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color)
{
    writeFillRect(x, y, w, h, color);
}

void Sipeed_ST7789::fillScreen(uint16_t color)
{
    writeFillRect(0, 0, _width, _height, color);
}

void Sipeed_ST7789::drawRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color)
{
    writeFastVLine(x, y, h, color);
    writeFastVLine(x + w, y, h, color);
    writeFastHLine(x, y, w, color);
    writeFastHLine(x, y + h, w, color);
}

void Sipeed_ST7789::drawRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color, uint16_t lineWidth)
{
    lcd_draw_rectangle(x, y, x+w-1, y+h-1, lineWidth, color);
}


uint16_t getValueByRotation(uint8_t rotation)
{
    uint16_t v = DIR_YX_RLDU;
    switch(rotation) {
        case 0:
            v = DIR_YX_RLDU;
            break;
        case 1:
            v = DIR_XY_RLUD;
            break;
        case 2:
            v = DIR_YX_LRUD;
            break;
        case 3:
            v = DIR_XY_LRDU;
            break;
    }
    return v;
}

/**************************************************************************/
/*!
    @brief      Set rotation setting for display
    @param  x   0 thru 3 corresponding to 4 cardinal rotations
*/
/**************************************************************************/
void Sipeed_ST7789::setRotation(uint8_t x) {
    rotation = (x & 3);
    configASSERT(rotation >= 0 && rotation <= 3);

    switch(rotation) {
        case 0:
        case 2:
            _width  = WIDTH;
            _height = HEIGHT;
            break;
        case 1:
        case 3:
            _width  = HEIGHT;
            _height = WIDTH;
            break;
    }
    _screenDir = getValueByRotation(rotation);
    lcd_set_direction((lcd_dir_t)_screenDir);
}

void Sipeed_ST7789::invertDisplay(boolean invert) {
    uint16_t _screenDir = getValueByRotation(rotation);
    if( invert )
    {
        switch(rotation) {
            case 0:
                _screenDir = DIR_YX_RLUD;
                break;
            case 1:
                _screenDir = DIR_XY_LRUD;
                break;
            case 2:
                _screenDir = DIR_YX_LRDU;
                break;
            case 3:
                _screenDir = DIR_XY_RLDU;
                break;
        }
    }
    lcd_set_direction((lcd_dir_t)_screenDir);
}

void Sipeed_ST7789::drawImage(uint16_t x, uint16_t y, uint16_t w, uint16_t h, uint16_t* img)
{
	int inside;
	int bufn, bufy, bufs;
	const uint16_t *tmppt;
	uint8_t *buf;
	uint16_t val;

	if ((x+w)<0 || x>=_width || (y+h)<0 || y>=_height)
		return;

	if (x<0 || (x+w)>=_width || y<0 || (y+h)>=_height)
		inside=0;
	else
		inside=1;

	tmppt=img;
	bufs = _height / 4;
	if (inside==1){
		for(int16_t j=0; j<h; j++, y++) {
			bufn=y / bufs;
			bufy=y % bufs;
			buf=&buffer[bufn][(bufy*_width + x) * 2];
			for(int16_t i=0; i<w; i++ ) {
				val=*(tmppt++);
				*(buf++)=val & 0xFF;
				*(buf++)=val >> 8;
			}
		}
	}else{
		for(int16_t j=0; j<h; j++, y++) {
			for(int16_t i=0; i<w; i++ ) {
				writePixel(x+i, y, *(tmppt++));
			}
		}
	}
}

